package cz.muni.fi.pv112.cv3;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.joml.Matrix3f;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import org.lwjgl.BufferUtils;
import org.lwjgl.Version;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.MemoryStack;
import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;
import static org.lwjgl.system.MemoryStack.stackPush;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Cv3 {

    private static final int SIZEOF_AXES_VERTEX = 6 * Float.BYTES;
    private static final int COLOR_OFFSET = 3 * Float.BYTES;

    private static final float AXES[] = {
        // .. position .......... color ....
        // x axis
        1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        // y axis
        0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        // z axis
        0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f
    };

    private static final int SIZEOF_MODEL_VERTEX = 6 * Float.BYTES;
    private static final int NORMAL_OFFSET = 3 * Float.BYTES;

    private Camera camera;

    // the window handle
    private long window;

    // window size
    private int width;
    private int height;
    private boolean resized = false;

    // animation
    private boolean animate = false;
    private float t = 0f;

    // rendering mode
    private int mode = GL_FILL;

    // model
    private ObjLoader teapot;

    // our OpenGL resources
    private int axesBuffer;
    private int teapotBuffer;
    private int axesArray;
    private int teapotArray;

    // our GLSL resources
    private int axesProgram;
    private int axesAspectUniformLoc;
    private int axesLengthUniformLoc;
    private int axesMvpUniformLoc;

    private int modelProgram;
    private int modelMvpUniformLoc;
    private int modelNUniformLoc;
    private int modelModelLoc;

    private int lightPositionLoc;
    private int lightAmbientColorLoc;
    private int lightDiffuseColorLoc;
    private int lightSpecularColorLoc;

    private int materialAmbientColorLoc;
    private int materialDiffuseColorLoc;
    private int materialSpecularColorLoc;
    private int materialShininessLoc;

    private int eyePositionLoc;

    public static void main(String[] args) {
        new Cv3().run();
    }

    public void run() {
        System.out.println("Hello LWJGL " + Version.getVersion() + "!");

        camera = new Camera();

        initGLFW();
        loop();

        // Free the window callbacks and destroy the window
        glfwFreeCallbacks(window);
        glfwDestroyWindow(window);

        // Terminate GLFW and free the error callback
        glfwTerminate();
        glfwSetErrorCallback(null).free();
    }

    private void initGLFW() {
        // Setup an error callback. The default implementation
        // will print the error message in System.err.
        GLFWErrorCallback.createPrint(System.err).set();

        // Initialize GLFW. Most GLFW functions will not work before doing this.
        if (!glfwInit()) {
            throw new IllegalStateException("Unable to initialize GLFW");
        }

        // Configure GLFW
        glfwDefaultWindowHints(); // optional, the current window hints are already the default
        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE); // the window will stay hidden after creation
        glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // the window will be resizable
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // request the OpenGL 3.3 core profile context
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

        // set initial width and height
        width = 640;
        height = 480;

        // Create the window
        window = glfwCreateWindow(width, height, "Hello World!", NULL, NULL);
        if (window == NULL) {
            throw new RuntimeException("Failed to create the GLFW window");
        }

        // Setup a key callback. It will be called every time a key is pressed, repeated or released.
        glfwSetKeyCallback(window, (window, key, scancode, action, mods) -> {
            if (action == GLFW_RELEASE) {
                switch (key) {
                    case GLFW_KEY_ESCAPE:
                        glfwSetWindowShouldClose(window, true); // We will detect this in the rendering loop
                        break;
                    case GLFW_KEY_A:
                        animate = !animate;
                        break;
                    case GLFW_KEY_T:
                        // TODO toggle fullscreen
                        break;
                    case GLFW_KEY_L:
                        mode = GL_LINE;
                        break;
                    case GLFW_KEY_F:
                        mode = GL_FILL;
                        break;
                }
            }
        });

        glfwSetMouseButtonCallback(window, (window, button, action, mods) -> {
            if (action == GLFW_PRESS) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, true);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, true);
                }
            } else if (action == GLFW_RELEASE) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, false);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, false);
                }
            }
        });

        glfwSetCursorPosCallback(window, (window, xpos, ypos) -> {
            camera.updateMousePosition(xpos, ypos);
        });

        // add window size callback
        glfwSetWindowSizeCallback(window, (window, width, height) -> {
            this.width = width;
            this.height = height;
            resized = true;
        });

        // Get the thread stack and push a new frame
        try (MemoryStack stack = stackPush()) {
            IntBuffer pWidth = stack.mallocInt(1); // int*
            IntBuffer pHeight = stack.mallocInt(1); // int*

            // Get the window size passed to glfwCreateWindow
            glfwGetWindowSize(window, pWidth, pHeight);

            // Get the resolution of the primary monitor
            GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());

            // Center the window
            glfwSetWindowPos(
                    window,
                    (vidmode.width() - pWidth.get(0)) / 2,
                    (vidmode.height() - pHeight.get(0)) / 2
            );
        } // the stack frame is popped automatically

        // Make the OpenGL context current
        glfwMakeContextCurrent(window);
        // Enable v-sync
        glfwSwapInterval(1);

        // Make the window visible
        glfwShowWindow(window);

        // This line is critical for LWJGL's interoperation with GLFW's
        // OpenGL context, or any context that is managed externally.
        // LWJGL detects the context that is current in the current thread,
        // creates the GLCapabilities instance and makes the OpenGL
        // bindings available for use.
        GL.createCapabilities();
    }

    private void loop() {
        // Prepare data for rendering
        init();

        // Run the rendering loop until the user has attempted to close
        // the window or has pressed the ESCAPE key.
        while (!glfwWindowShouldClose(window)) {
            render();

            glfwSwapBuffers(window); // swap the color buffers

            // Poll for window events. The key callback above will only be
            // invoked during this call.
            glfwPollEvents();
        }
    }

    private void init() {
        // empty scene color
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glLineWidth(3.0f); // makes lines thicker

        glEnable(GL_DEPTH_TEST);

        // load GLSL program (vertex and fragment shaders)
        try {
            axesProgram = loadProgram("/resources/shaders/axes.vs.glsl",
                    "/resources/shaders/axes.fs.glsl");
            modelProgram = loadProgram("/resources/shaders/model.vs.glsl",
                    "/resources/shaders/model.fs.glsl");
        } catch (IOException ex) {
            Logger.getLogger(Cv3.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }

        // get uniform locations
        // axes program uniforms
        axesAspectUniformLoc = glGetUniformLocation(axesProgram, "aspect");
        axesLengthUniformLoc = glGetUniformLocation(axesProgram, "len");
        axesMvpUniformLoc = glGetUniformLocation(axesProgram, "MVP");

        // model program uniforms
        modelMvpUniformLoc = glGetUniformLocation(modelProgram, "MVP");
        modelNUniformLoc = glGetUniformLocation(modelProgram, "N");
        modelModelLoc = glGetUniformLocation(modelProgram, "model");

        // Task 3: store uniform variable locations for light properties (lightPosition, lightAmbientColor, lightDiffuseColor)
        //           use the prepared lightPositionLoc, lightAmbientColorLoc and lightDiffuseColorLoc class variables 
        // Task 4: store uniform variable location for lightSpecularColor into lightSpecularColorLoc class variable

        // Task 4: store uniform variable location for eyePosition into eyePositionLoc class variable

        // Task 5: store uniform variable locations for material properties:
        //           uniforms are materialAmbientColor, materialDiffuseColor, materialSpecularColor, materialShininess              
        //           the names of class variable are analogous to the previous tasks

        // create buffers with geometry
        int[] buffers = new int[2];
        glGenBuffers(buffers);
        axesBuffer = buffers[0];
        teapotBuffer = buffers[1];

        // fill a buffers with geometry
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        glBufferData(GL_ARRAY_BUFFER, AXES, GL_STATIC_DRAW);

        // load teapot and fill buffer with teapot data
        teapot = new ObjLoader("/resources/models/teapot.obj");
        try {
            teapot.load();
        } catch (IOException ex) {
            Logger.getLogger(Cv3.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        int length = 3 * 6 * teapot.getTriangleCount();
        FloatBuffer teapot = BufferUtils.createFloatBuffer(length);
        for (int f = 0; f < this.teapot.getTriangleCount(); f++) {
            int[] pi = this.teapot.getVertexIndices().get(f);
            int[] ni = this.teapot.getNormalIndices().get(f);
            for (int i = 0; i < 3; i++) {
                float[] position = this.teapot.getVertices().get(pi[i]);
                float[] normal = this.teapot.getNormals().get(ni[i]);
                teapot.put(position);
                teapot.put(normal);
            }
        }
        teapot.rewind();
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glBufferData(GL_ARRAY_BUFFER, teapot, GL_STATIC_DRAW);

        // clear buffer binding, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        // create a vertex array object for the geometry
        int[] arrays = new int[2];
        glGenVertexArrays(arrays);
        axesArray = arrays[0];
        teapotArray = arrays[1];

        // get axes program attributes
        int positionAttribLoc = glGetAttribLocation(axesProgram, "position");
        int colorAttribLoc = glGetAttribLocation(axesProgram, "color");
        // bind axes buffer
        glBindVertexArray(axesArray);
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        // stride and offset are employed as both position and color data reside in the same vertex buffer
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, 0);
        glEnableVertexAttribArray(colorAttribLoc);
        glVertexAttribPointer(colorAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, COLOR_OFFSET);

        // get cube program attributes
        positionAttribLoc = glGetAttribLocation(modelProgram, "position");
        int normalAttribLoc = glGetAttribLocation(modelProgram, "normal");

        // bind teapot buffer
        glBindVertexArray(teapotArray);
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);

        // clear bindings, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }

    private void render() {
        // Resize OpenGL viewport, i.e., the (bitmap) extents to that is the
        // OpenGL screen space [-1, 1] mapped.
        if (resized) {
            glViewport(0, 0, width, height);
            resized = false;
        }

        // animate variables
        if (animate) {
            t += 0.02f;
        }

        glPolygonMode(GL_FRONT_AND_BACK, mode);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        Matrix4f projection = new Matrix4f()
                .perspective((float) Math.toRadians(60f), width / (float) height, 1, 500);

        Matrix4f view = new Matrix4f()
                .lookAt(camera.getEyePosition(), new Vector3f(0, 0, 0), new Vector3f(0, 1, 0));

        // Task 5: create instance of Material class and use it to draw model using drawModel(..., Material material)
        //           constructor parameters are Material(Vector3f ambient, Vector3f diffuse, Vector3f specular, float shininess)
        //           try different values for colors of your material
        // Task 6: use one of the material values from the attached PDF file
        drawModel(new Matrix4f(), view, projection, teapotArray, teapot.getTriangleCount() * 3);

        // Task 7: draw two more teapots
        //           transform them to [-5, 0, 0] and [5, 0, 0]
        //           create two more materials by using values from the attached PDF file and use those to draw these three teapots
    }

    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int count) {
        drawModel(model, view, projection, vao, count, null);
    }

    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int count, Material material) {
        // compute model-view-projection matrix
        Matrix4f mvp = new Matrix4f(projection)
                .mul(view)
                .mul(model);

        // compute normal matrix
        Matrix3f n = model.get3x3(new Matrix3f())
                .invert()
                .transpose();

        glUseProgram(modelProgram);
        glBindVertexArray(vao); // bind vertex array to draw

        // Task 3: set values for lightPosition, lightAmbientColor and lightDiffuseColor uniform variables using glUniform(3|4)f functions
        //           set position to (2, 3, 1, 0)
        //           set ambient color  to (0.15f, 0.15f, 0.15f)
        //           set diffuse color to (1, 1, 1)
        // Task 4: set the value of lightSpecularColor uniform value to pure red light (1, 0, 0)
        // Task 6: change the light diffuse and specular colors to pure white
        // Task 8: change the position to (0, 5, 0, 1) to set the light as a point light
        //           try different positions and switch between 0 and 1 in the w compontent, to look at differences between the lights

        // Task 4: set eyePosition uniform variable to camera's position using glUniform3f(int location, float x, float y, float z)
        //           you can get the current position from camera.getEyePosition(), note it is a Vector3f and glUniform3f takes 3 floats

        // Task 5: set values for materialAmbientColor, materialDiffuseColor, materialSpecularColor and materialShininess uniform variables
        //           using the material variable passed to drawModel() function


        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        FloatBuffer nData = BufferUtils.createFloatBuffer(9);
        FloatBuffer modelData = BufferUtils.createFloatBuffer(16);
        mvp.get(mvpData);
        n.get(nData);
        model.get(modelData);
        glUniformMatrix4fv(modelMvpUniformLoc, false, mvpData); // pass MVP matrix to shader
        glUniformMatrix3fv(modelNUniformLoc, false, nData); // pass Normal matrix to shader
        glUniformMatrix4fv(modelModelLoc, false, modelData); // pass model matrix to shader

        glDrawArrays(GL_TRIANGLES, 0, count);

        glBindVertexArray(0);
        glUseProgram(0);
    }

    private void drawAxes(Matrix4f modelViewProjection, float length) {
        glUseProgram(axesProgram);
        glBindVertexArray(axesArray);

        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        modelViewProjection.get(mvpData);
        glUniform1f(axesLengthUniformLoc, length);
        glUniform1f(axesAspectUniformLoc, width / (float) height);
        glUniformMatrix4fv(axesMvpUniformLoc, false, mvpData);

        glDrawArrays(GL_LINES, 0, 6);

        glBindVertexArray(0);
        glUseProgram(0);
    }

    private int loadShader(String filename, int shaderType) throws IOException {
        String source = readAllFromResource(filename);
        int shader = glCreateShader(shaderType);

        // create and compile GLSL shader
        glShaderSource(shader, source);
        glCompileShader(shader);

        // check GLSL shader compile status
        int status = glGetShaderi(shader, GL_COMPILE_STATUS);
        if (status == GL_FALSE) {
            String log = glGetShaderInfoLog(shader);
            System.err.println(log);
        }

        return shader;
    }

    private int loadProgram(String vertexShaderFile, String fragmentShaderFile) throws IOException {
        // load vertex and fragment shaders (GLSL)
        int vs = loadShader(vertexShaderFile, GL_VERTEX_SHADER);
        int fs = loadShader(fragmentShaderFile, GL_FRAGMENT_SHADER);

        // create GLSL program, attach shaders and compile it
        int program = glCreateProgram();
        glAttachShader(program, vs);
        glAttachShader(program, fs);
        glLinkProgram(program);

        int status = glGetProgrami(program, GL_LINK_STATUS);
        if (status == GL_FALSE) {
            String log = glGetProgramInfoLog(program);
            System.err.println(log);
        }

        return program;
    }

    private String readAllFromResource(String resource) throws IOException {
        InputStream is = Cv3.class.getResourceAsStream(resource);
        if (is == null) {
            throw new IOException("Resource not found: " + resource);
        }

        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        StringBuilder sb = new StringBuilder();

        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }

        return sb.toString();
    }

}
